==> kaffe-0.9.0.tgz: kaffe-0.9.0/kaffe/kaffevm/findInJar.c <==
/*
 * findInJar.c
 * Search the CLASSPATH for the given class or property name.
 *
 * Copyright (c) 1996 T. J. Wilkinson & Associates, London, UK.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * Written by Tim Wilkinson <tim@tjwassoc.demon.co.uk>, 1996.
 */

#define	FDBG(s)
#define ZDBG(s)
#define	PDBG(s)
#define	CDBG(s)

#include "config.h"
#include "config-std.h"
#include "config-io.h"
#include "config-mem.h"
#include "gtypes.h"
#include "file.h"
#include "exception.h"
#include "zipfile.h"
#include "readClass.h"
#include "paths.h"
#include "flags.h"
#include "errors.h"
#include "lerrno.h"
#include "locks.h"

#define	CLASSPATH1	"KAFFE_CLASSPATH"
#define	CLASSPATH2	"CLASSPATH"
#define	MAXBUF		256
#define	MAXPATHELEM	16

#define	CP_INVALID	0
#define	CP_FILE		1
#define	CP_DIR		2

static struct {
	int	type;
	char*	path;
	ZipFile	zipf;
} classpath[MAXPATHELEM+1];

char* realClassPath;

void initClasspath(void);
classFile findInJar(char*);

/*
 * Find the named class in a directory or zip file.
 */
Hjava_lang_Class*
findClass(Hjava_lang_Class* class, char* cname)
{
	char buf[MAXBUF];
	classFile hand;

	/* Look for the class */
CDBG(	printf("Scanning for class %s\n", cname);		)

	strcpy(buf, cname);
	strcat(buf, ".class");

	/* Find class in Jar file */
	hand = findInJar(buf);
	if (hand.size > 0) {
		if (flag_classload) {
			fprintf(stderr, "Loading class '%s'.\n", cname);
		}
		class = readClass(class, &hand, NULL);
		if (hand.base != 0) {
			checked_free(hand.base);
		}
		return (class);
	}

	/*
	 * Certain classes are essential.  If we don't find them then
	 * abort.
	 */
	if (strcmp(cname, "java/lang/ClassNotFoundException") == 0 ||
	    strcmp(cname, "java/lang/Object") == 0) {
		fprintf(stderr, "Cannot find essential class '%s' in class library ... aborting.\n", cname);
		abort();
	}
	return (0);
}

/*
 * Locate the given name in the CLASSPATH.
 */
classFile
findInJar(char* cname)
{
	static quickLock jarlock;
	int i;
	char buf[MAXBUF];
	int fp;
	struct stat sbuf;
	classFile hand;
	ZipDirectory *zipd;
	int j;

	/* Look for the class */
CDBG(	printf("Scanning for element %s\n", cname);		)

	/* One into the jar at once */
	lockMutex(&jarlock);

	hand.size = -1;
	for (i = 0; classpath[i].path != 0; i++) {
		switch (classpath[i].type) {
		case CP_FILE:
ZDBG(			printf("Opening zip file %s for %s\n", classpath[i].path, cname); )
			if (classpath[i].zipf.central_directory == 0) {
				classpath[i].zipf.fd = open(classpath[i].path, O_RDONLY|O_BINARY);
				if (classpath[i].zipf.fd < 0 || read_zip_archive (&classpath[i].zipf) != 0) {
					continue;
				}
				close(classpath[i].zipf.fd);
			}

			zipd = (ZipDirectory*)classpath[i].zipf.central_directory;
			for (j = 0; j < classpath[i].zipf.count; j++, zipd = ZIPDIR_NEXT(zipd)) {

ZDBG(				printf ("%d: size:%d, name(#%d)%s, offset:%d\n", i, zipd->size, zipd->filename_length, ZIPDIR_FILENAME (zipd), zipd->filestart);  )
				if (cname[0] == ZIPDIR_FILENAME(zipd)[0] && strcmp(cname, ZIPDIR_FILENAME(zipd)) == 0) {

ZDBG(					printf("FOUND!!\n");		)

#if defined(HAVE_MMAP) && defined(HAVE_UNALIGNEDACCESS)
					hand.size = zipd->size;
					hand.base = NULL;
					hand.buf = classpath[i].zipf.mmap_base + zipd->filestart;
					goto mmap_found;
#else
					fp = open(classpath[i].path, O_RDONLY|O_BINARY);
					lseek(fp, zipd->filestart, SEEK_SET);
					hand.size = zipd->size;
					goto found;
#endif
				}
			}
			continue;

		case CP_DIR:
			strcpy(buf, classpath[i].path);
			strcat(buf, DIRSEP);
			strcat(buf, cname);
FDBG(			printf("Opening java file %s for %s\n", buf, cname); )
			fp = open(buf, O_RDONLY|O_BINARY);
			if (fp < 0 || fstat(fp, &sbuf) < 0) {
				continue;
			}
			hand.size = sbuf.st_size;

		found:
			hand.base = checked_malloc(hand.size);
			hand.buf = hand.base;

			i = 0;
			while (i < hand.size) {
				j = read(fp, hand.buf, hand.size - i);
				if (j >= 0) {
					i += j;
				}
				else if (!(errno == EAGAIN || errno == EWOULDBLOCK || errno == EINTR)) {
					throwException(IOException);
				}
			}
			close(fp);
#if defined(HAVE_MMAP) && defined(HAVE_UNALIGNEDACCESS)
		mmap_found:
#endif
			break;


		/* Ignore bad entries */
		default:
			continue;
		}
		break;
	}

	unlockMutex(&jarlock);

	return (hand);
}

/*
 * Initialise class path.
 */
void
initClasspath(void)
{
	struct stat sbuf;
	char* cp;
	int i;

	/* If classpath isn't set, get if from the environment. */
	if (realClassPath == 0) {
		cp = getenv(CLASSPATH1);
		if (cp == 0) {
			cp = getenv(CLASSPATH2);
		}
		if (cp == 0) {
#if defined(DEFAULT_CLASSPATH)
			cp = DEFAULT_CLASSPATH;
#else
			fprintf(stderr, "CLASSPATH is not set!\n");
			exit(1);
#endif
		}
		realClassPath = checked_malloc(strlen(cp) + 1);
		strcpy(realClassPath, cp);
	}
	cp = realClassPath;

PDBG(	printf("initClasspath(): '%s'\n", cp);				)
	for (i = 0; cp != 0 && i < MAXPATHELEM; i++) {
		classpath[i].path = cp;
		cp = strchr(cp, PATHSEP);
		if (cp != 0) {
			*cp = 0;
			cp++;
		}
		if (stat(classpath[i].path, &sbuf) < 0) {
			classpath[i].type = CP_INVALID;
		}
		else if (S_ISDIR(sbuf.st_mode)) {
			classpath[i].type = CP_DIR;
		}
		else {
			classpath[i].type = CP_FILE;
		}
PDBG(		printf("path '%s' type %d\n", classpath[i].path, classpath[i].type); )
	}
	i++;
	classpath[i].path = 0;
}

/*
 * Add an entry in the Classpath dynamically.
 */
int
addClasspath(char* cp)
{
	struct stat sbuf;
	int i;

	/* If classpath isn't set, initialize it */
	if (realClassPath == 0) {
		initClasspath();
	}

PDBG(	printf("addClasspath(): '%s'\n", cp);				)

	for (i = 0; i < MAXPATHELEM; i++) {
		if (classpath[i].path == 0) {
			goto found;
		}
		if (!strcmp(cp, classpath[i].path)) {
			return(0); /* already in */
		}
	}
	fprintf(stderr, "addClasspath : CLASSPATH is full!\n");
	return(-1);

	found:
	classpath[i].path = strdup(cp);
	if (stat(classpath[i].path, &sbuf) < 0) {
		classpath[i].type = CP_INVALID;
	}
	else if (S_ISDIR(sbuf.st_mode)) {
		classpath[i].type = CP_DIR;
	}
	else {
		classpath[i].type = CP_FILE;
	}
PDBG(	printf("path '%s' type %d\n", classpath[i].path, classpath[i].type); )
	i++;
	classpath[i].path = 0;
}

==> kaffe-0.9.1.tgz: kaffe-0.9.1/kaffe/kaffevm/findInJar.c <==
/*
 * findInJar.c
 * Search the CLASSPATH for the given class or property name.
 *
 * Copyright (c) 1996 T. J. Wilkinson & Associates, London, UK.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * Written by Tim Wilkinson <tim@tjwassoc.demon.co.uk>, 1996.
 */

#define	FDBG(s)
#define ZDBG(s)
#define	PDBG(s)
#define	CDBG(s)

#include "config.h"
#include "config-std.h"
#include "config-io.h"
#include "config-mem.h"
#include "gtypes.h"
#include "file.h"
#include "exception.h"
#include "zipfile.h"
#include "readClass.h"
#include "paths.h"
#include "flags.h"
#include "errors.h"
#include "lerrno.h"
#include "locks.h"

#define	CLASSPATH1	"KAFFE_CLASSPATH"
#define	CLASSPATH2	"CLASSPATH"
#define	MAXBUF		256
#define	MAXPATHELEM	16

#define	CP_INVALID	0
#define	CP_FILE		1
#define	CP_DIR		2

static struct {
	int	type;
	char*	path;
	ZipFile	zipf;
} classpath[MAXPATHELEM+1];

char* realClassPath;

void initClasspath(void);
classFile findInJar(char*);

/*
 * Find the named class in a directory or zip file.
 */
Hjava_lang_Class*
findClass(Hjava_lang_Class* class, char* cname)
{
	char buf[MAXBUF];
	classFile hand;

	/* Look for the class */
CDBG(	printf("Scanning for class %s\n", cname);		)

	strcpy(buf, cname);
	strcat(buf, ".class");

	/* Find class in Jar file */
	hand = findInJar(buf);
	if (hand.size > 0) {
		if (flag_classload) {
			fprintf(stderr, "Loading class '%s'.\n", cname);
		}
		class = readClass(class, &hand, NULL);
		if (hand.base != 0) {
			gc_free_fixed(hand.base);
		}
		return (class);
	}

	/*
	 * Certain classes are essential.  If we don't find them then
	 * abort.
	 */
	if (strcmp(cname, "java/lang/ClassNotFoundException") == 0 ||
	    strcmp(cname, "java/lang/Object") == 0) {
		fprintf(stderr, "Cannot find essential class '%s' in class library ... aborting.\n", cname);
		abort();
	}
	return (0);
}

/*
 * Locate the given name in the CLASSPATH.
 */
classFile
findInJar(char* cname)
{
	static quickLock jarlock;
	int i;
	char buf[MAXBUF];
	int fp;
	struct stat sbuf;
	classFile hand;
	ZipDirectory *zipd;
	int j;

	/* Look for the class */
CDBG(	printf("Scanning for element %s\n", cname);		)

	/* One into the jar at once */
	lockMutex(&jarlock);

	hand.size = -1;
	for (i = 0; classpath[i].path != 0; i++) {
		switch (classpath[i].type) {
		case CP_FILE:
ZDBG(			printf("Opening zip file %s for %s\n", classpath[i].path, cname); )
			if (classpath[i].zipf.central_directory == 0) {
				classpath[i].zipf.fd = open(classpath[i].path, O_RDONLY|O_BINARY);
				if (classpath[i].zipf.fd < 0 || read_zip_archive (&classpath[i].zipf) != 0) {
					continue;
				}
				close(classpath[i].zipf.fd);
			}

			zipd = (ZipDirectory*)classpath[i].zipf.central_directory;
			for (j = 0; j < classpath[i].zipf.count; j++, zipd = ZIPDIR_NEXT(zipd)) {

ZDBG(				printf ("%d: size:%d, name(#%d)%s, offset:%d\n", i, zipd->size, zipd->filename_length, ZIPDIR_FILENAME (zipd), zipd->filestart);  )
				if (cname[0] == ZIPDIR_FILENAME(zipd)[0] && strcmp(cname, ZIPDIR_FILENAME(zipd)) == 0) {

ZDBG(					printf("FOUND!!\n");		)

					/* If size is negative then this
					 * is compressed !!
					 */
					if (zipd->size == -1) {
						throwException(IOException("cannot handle compressed entry in JAR/ZIP file"));
					}

#if defined(HAVE_MMAP) && defined(HAVE_UNALIGNEDACCESS)
					hand.size = zipd->size;
					hand.base = NULL;
					hand.buf = classpath[i].zipf.mmap_base + zipd->filestart;
					goto mmap_found;
#else
					fp = open(classpath[i].path, O_RDONLY|O_BINARY);
					lseek(fp, zipd->filestart, SEEK_SET);
					hand.size = zipd->size;
					goto found;
#endif
				}
			}
			continue;

		case CP_DIR:
			strcpy(buf, classpath[i].path);
			strcat(buf, DIRSEP);
			strcat(buf, cname);
FDBG(			printf("Opening java file %s for %s\n", buf, cname); )
			fp = open(buf, O_RDONLY|O_BINARY);
			if (fp < 0 || fstat(fp, &sbuf) < 0) {
				continue;
			}
			hand.size = sbuf.st_size;

		found:
			hand.base = gc_malloc_fixed(hand.size);
			hand.buf = hand.base;

			i = 0;
			while (i < hand.size) {
				j = read(fp, hand.buf, hand.size - i);
				if (j >= 0) {
					i += j;
				}
				else if (!(errno == EAGAIN || errno == EWOULDBLOCK || errno == EINTR)) {
					throwException(IOException("failed to read JAR data"));
				}
			}
			close(fp);
#if defined(HAVE_MMAP) && defined(HAVE_UNALIGNEDACCESS)
		mmap_found:
#endif
			break;


		/* Ignore bad entries */
		default:
			continue;
		}
		break;
	}

	unlockMutex(&jarlock);

	return (hand);
}

/*
 * Initialise class path.
 */
void
initClasspath(void)
{
	struct stat sbuf;
	char* cp;
	int i;

	/* If classpath isn't set, get if from the environment. */
	if (realClassPath == 0) {
		cp = getenv(CLASSPATH1);
		if (cp == 0) {
			cp = getenv(CLASSPATH2);
		}
		if (cp == 0) {
#if defined(DEFAULT_CLASSPATH)
			cp = DEFAULT_CLASSPATH;
#else
			fprintf(stderr, "CLASSPATH is not set!\n");
			exit(1);
#endif
		}
		realClassPath = gc_malloc_fixed(strlen(cp) + 1);
		strcpy(realClassPath, cp);
	}
	cp = realClassPath;

PDBG(	printf("initClasspath(): '%s'\n", cp);				)
	for (i = 0; cp != 0 && i < MAXPATHELEM; i++) {
		classpath[i].path = cp;
		cp = strchr(cp, PATHSEP);
		if (cp != 0) {
			*cp = 0;
			cp++;
		}
		if (stat(classpath[i].path, &sbuf) < 0) {
			classpath[i].type = CP_INVALID;
		}
		else if (S_ISDIR(sbuf.st_mode)) {
			classpath[i].type = CP_DIR;
		}
		else {
			classpath[i].type = CP_FILE;
		}
PDBG(		printf("path '%s' type %d\n", classpath[i].path, classpath[i].type); )
	}
	i++;
	classpath[i].path = 0;
}

/*
 * Add an entry in the Classpath dynamically.
 */
int
addClasspath(char* cp)
{
	struct stat sbuf;
	int i;

	/* If classpath isn't set, initialize it */
	if (realClassPath == 0) {
		initClasspath();
	}

PDBG(	printf("addClasspath(): '%s'\n", cp);				)

	for (i = 0; i < MAXPATHELEM; i++) {
		if (classpath[i].path == 0) {
			goto found;
		}
		if (!strcmp(cp, classpath[i].path)) {
			return(0); /* already in */
		}
	}
	fprintf(stderr, "addClasspath : CLASSPATH is full!\n");
	return(-1);

	found:
	classpath[i].path = strdup(cp);
	if (stat(classpath[i].path, &sbuf) < 0) {
		classpath[i].type = CP_INVALID;
	}
	else if (S_ISDIR(sbuf.st_mode)) {
		classpath[i].type = CP_DIR;
	}
	else {
		classpath[i].type = CP_FILE;
	}
PDBG(	printf("path '%s' type %d\n", classpath[i].path, classpath[i].type); )
	i++;
	classpath[i].path = 0;
	return(1);
}
